﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.SymbolicExpressionTreeEncoding {
  [Item("ChangeNodeTypeMoveMaker", "Peforms a change node type move on a given symbolic expression tree and updates the quality.")]
  [StorableClass]
  public class ChangeNodeTypeMoveMaker : SingleSuccessorOperator, ISymbolicExpressionTreeMoveOperator, IMoveMaker {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<DoubleValue> MoveQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }
    public ILookupParameter<ChangeNodeTypeMove> ChangeNodeTypeMoveParameter {
      get { return (ILookupParameter<ChangeNodeTypeMove>)Parameters["ChangeNodeTypeMove"]; }
    }
    public ILookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (ILookupParameter<ISymbolicExpressionTree>)Parameters["SymbolicExpressionTree"]; }
    }

    [StorableConstructor]
    protected ChangeNodeTypeMoveMaker(bool deserializing) : base(deserializing) { }
    protected ChangeNodeTypeMoveMaker(ChangeNodeTypeMoveMaker original, Cloner cloner) : base(original, cloner) { }
    public ChangeNodeTypeMoveMaker()
      : base() {
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The quality of the solution."));
      Parameters.Add(new LookupParameter<ChangeNodeTypeMove>("ChangeNodeTypeMove", "The move to evaluate."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The relative quality of the move."));
      Parameters.Add(new LookupParameter<ISymbolicExpressionTree>("SymbolicExpressionTree", "The symbolic expression tree on which the move should be applied."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ChangeNodeTypeMoveMaker(this, cloner);
    }

    public override IOperation Apply() {
      ChangeNodeTypeMove move = ChangeNodeTypeMoveParameter.ActualValue;
      var tree = SymbolicExpressionTreeParameter.ActualValue;
      DoubleValue moveQuality = MoveQualityParameter.ActualValue;
      DoubleValue quality = QualityParameter.ActualValue;

      var newNode = move.NewChild;
      var oldNode = move.Parent.GetSubtree(move.SubtreeIndex);
      var children = new List<ISymbolicExpressionTreeNode>(oldNode.Subtrees);
      while (oldNode.SubtreeCount > 0) oldNode.RemoveSubtree(0);
      foreach (var child in children) {
        newNode.AddSubtree(child);
      }
      move.Parent.RemoveSubtree(move.SubtreeIndex);
      move.Parent.InsertSubtree(move.SubtreeIndex, newNode);

      quality.Value = moveQuality.Value;

      return base.Apply();
    }
  }
}
